import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.*;

import sv.kernel.*;
import sv.ui.*;
import sv.util.*;
import sv.drawer.*;

/** 
 * GCPOneD.java
 * Define a canvas to draw the graph
 * @version 1.0    10/15/97
 * @author Byeongseob Ki
 */

class GCPOneD extends GCPlane implements KeyListener, MouseMotionListener {

  boolean isLegendMove = false;
  Rectangle legendCenter;
 
  // closest point
  private boolean drawClosest = false;
  private float[] closest;
  private String displayStr;

  /**
   * constructor
   */
  public GCPOneD(GFPOneD gframe) {
     super(gframe);

     addKeyListener(this);
     addMouseMotionListener(this);
     enableEvents(AWTEvent.MOUSE_EVENT_MASK);
  }  


  /**
   * overrided paint method
   */
  public void paint(Graphics g) {

    paintDone = false;
    Dimension d = getSize();

    // virtual desktop
    Rectangle r = gframe.getBounds();
    if (!r.equals(rect)) {
       if (svServer.controller != null)
          svServer.controller.desktopRepaint();
    }
    
    if ((offScrImage == null) || 
          (d.width != canvasSize.width) || (d.height != canvasSize.height)) {
       offScrImage = createImage(d.width, d.height);
       offScrGc = offScrImage.getGraphics();
       changeNotify(d);
    }

    printing = g instanceof PrintGraphics;
      
    if (printing) {

      // modifies variables
      Graphics savedgc = offScrGc;
      offScrGc = g;
      
      Dimension pagedimension = ((PrintGraphics)g).getPrintJob().getPageDimension();
      
      printwidth = pagedimension.width;
      printheight = canvasSize.height * canvasSize.width / canvasSize.width;

      if (printheight > pagedimension.height) {
        printheight = pagedimension.height;
        printwidth = canvasSize.width * canvasSize.height / canvasSize.height;
      }

      // erase whole canvas
      offScrGc.setColor(gframe.config.bgColor);
      offScrGc.fillRect(0, 0, printwidth, printheight);

      // draw point data 
      drawPData(offScrGc);

      // draw attached labels and title
      drawLabels(offScrGc);
      objectsDraw(offScrGc);

      offScrGc = savedgc;
      paintDone = true;
      return;
    }
    

    // erase whole canvas
    offScrGc.setColor(gframe.config.bgColor);
    offScrGc.fillRect(0, 0, d.width, d.height);

    if ((getSize().width < 50) || (getSize().height < 50)) {
       offScrGc.setColor(Color.blue);
       offScrGc.fillRect(0, 0, d.width, d.height);
       offScrGc.setColor(Color.red);
       offScrGc.drawString("Make this larger ...", 5, (int)(getSize().height/2));
       g.drawImage(offScrImage, 0, 0, this);
       paintDone = true;
       return;
    }

    // draw point data 
    drawPData(offScrGc);

    // draw attached labels and title
    drawLabels(offScrGc);

    if (svServer.mergeHandler.mergeMode && gframe.mergeSelected) {
       Font savedFont = offScrGc.getFont();
       offScrGc.setColor(Color.red);
       offScrGc.setFont(new Font("Helvetica", Font.BOLD, 24));

       offScrGc.drawString("a" + gframe.mergeIndex, d.width/2, d.height/2);
       offScrGc.setFont(savedFont);
    }

    // draw new objects
    objectsDraw(offScrGc);

    g.drawImage(offScrImage, 0, 0, this);
    paintDone = true;
  }

  /**
   * overrided update method (we need this function for remove flicker)
   */
  public void update(Graphics g) {
    paint(g);
  }

  // implementation of changeNotify
  protected void changeNotify(Dimension d) {
    canvasSize = d;

    float xresize = (float)d.width / 390f;
    float yresize = (float)d.height / 300f;
    if (xresize < yresize) resizeFactor = xresize;
    else                   resizeFactor = yresize;

    if (offScrImage == null) return;

    fontSize = (int)(12.0f * resizeFactor); 
    if (fontSize > 24) fontSize = 24;

    offScrGc.setFont(new Font("TimesRoman", Font.PLAIN, fontSize));
    FontMetrics fm = offScrGc.getFontMetrics();
    int tickhgt = fm.getHeight();
    int lblhgt = 20;

    // subtract the parts for labels and title to calculate PDataArea size
    if (gframe.config.tickmark) {

      //////////////////////////////////////////////
      // ticker and label
      if (gframe.config.display_label) {
          pda = new Rectangle(60+lblhgt, 20+tickhgt, 
                d.width-lblhgt-80, d.height-40-2*tickhgt-lblhgt);

          gframe.config.xLabelPoint.x = (pda.x + pda.width) / 2;
          gframe.config.xLabelPoint.y = d.height - lblhgt - 10;
          gframe.config.yLabelPoint.x = 10;
          gframe.config.yLabelPoint.y = (pda.y + pda.height) / 2;

      //////////////////////////////////////////////
      // ticker only
      } else {                             
          pda = new Rectangle(60, 20+tickhgt, 
                d.width-80, d.height-40-2*tickhgt);
      }
      axis.setTickMode(true);

    } else {

      //////////////////////////////////////////////
      // label only
      if (gframe.config.display_label) {
          pda = new Rectangle(20+lblhgt, 20+ tickhgt, 
                d.width-lblhgt-40, d.height-40-2*tickhgt-lblhgt);

          gframe.config.xLabelPoint.x = (pda.x + pda.width) / 2;
          gframe.config.xLabelPoint.y = d.height - lblhgt - 10;
          gframe.config.yLabelPoint.x = 10;
          gframe.config.yLabelPoint.y = (pda.y + pda.height) / 2;

      //////////////////////////////////////////////
      // no label, no ticker
      } else {                           
          pda = new Rectangle(20, 20+tickhgt,
                 d.width-40, d.height-40-2*tickhgt);
      }
      axis.setTickMode(false);
    }

    if (gframe.config.display_legend) {
      pda = new Rectangle(pda.x, pda.y+4*tickhgt,
                  pda.width, pda.height-4*tickhgt);
      legendCenter = null;
    }
    change_pdaCorners();

    // SHOULD SEND NOTIFICATION TO ALL WHITEBOARD OBJECT
    gframe.changeWinSize(d);
  }



  /////////////////////////////////////////////////////////////
  // draw labels and title, also dragging objects
  /////////////////////////////////////////////////////////////
  private void drawLabels(Graphics g) {
    Font saved = g.getFont();

    if (isDragged) {
        if (drawingmode)
            g.setColor(Color.blue);
        else
            g.setColor(Color.black);

        int stype = svServer.shapes.getShapeType();
        if (drawingmode && (stype == ShapeProperties.LINE ||
                       stype == ShapeProperties.MULTILINE ||
                       stype == ShapeProperties.BEZIER)) {
            g.drawLine(dragStartX, dragStartY, dragEndX, dragEndY);
        
        } else if (drawingmode && stype == ShapeProperties.POLYGON) {
            if (mPointVec == null) return;
          
            int togo = mPointVec.size() - 1;
            Point p1, p2;

            if (togo > 1) {
               for (int i=0; i<togo; i++) {
                  p1 = (Point)mPointVec.elementAt(i);
                  p2 = (Point)mPointVec.elementAt(i+1);
                  g.drawLine(p1.x, p1.y, p2.x, p2.y);
	       }
   	    }

        } else if (drawingmode && stype == ShapeProperties.MOUSE) {
            if (mPointVec == null) return;
          
            int togo = mPointVec.size() - 1;
            Point p1, p2;

            if (togo < 1) return;

            switch (svServer.shapes.getLineType()) {
              case LineShape.NOARROW_LINE:
              case LineShape.RIGHTFILLARROW_LINE:
              case LineShape.RIGHTARROW_LINE:
              case LineShape.LEFTFILLARROW_LINE:
              case LineShape.LEFTARROW_LINE:
              case LineShape.LRFILLARROW_LINE:
              case LineShape.LRARROW_LINE:

		for (int i=0; i<togo; i++) {
                  p1 = (Point)mPointVec.elementAt(i);
                  p2 = (Point)mPointVec.elementAt(i+1);
                  SVUIC.drawLine(g, p1.x, p1.y, p2.x, p2.y, 
                                 svServer.shapes.getLineWidth(), 
                                 svServer.shapes.getLineColor());
                }
                break;

             case LineShape.NOARROW_DLINE:
             case LineShape.RIGHTFILLARROW_DLINE:
             case LineShape.RIGHTARROW_DLINE:
             case LineShape.LEFTFILLARROW_DLINE:
             case LineShape.LEFTARROW_DLINE:
             case LineShape.LRFILLARROW_DLINE:
             case LineShape.LRARROW_DLINE:

		for (int i=0; i<togo; i++) {
                  p1 = (Point)mPointVec.elementAt(i);
                  p2 = (Point)mPointVec.elementAt(i+1);
                  SVUIC.drawDottedLine(g, p1.x, p1.y, p2.x, p2.y, 
                                 svServer.shapes.getLineWidth(), 
                                 svServer.shapes.getDashSize(), 
                                 svServer.shapes.getLineColor());
                }
                break;
            }

        } else {
           if (inversedX) {
              if (inversedY)
                g.drawRect(dragEndX, dragEndY, 
                          dragStartX - dragEndX, dragStartY - dragEndY);
              else
                g.drawRect(dragEndX, dragStartY, 
                          dragStartX - dragEndX, dragEndY - dragStartY);
            } else {
              if (inversedY)
                 g.drawRect(dragStartX, dragEndY, 
                       dragEndX - dragStartX, dragStartY - dragEndY);
              else
                 g.drawRect(dragStartX, dragStartY, 
                        dragEndX - dragStartX, dragEndY - dragStartY);
	    }
	}
    }

    if (gframe.config.display_label) {
       int hgt = g.getFontMetrics().getHeight();

       text.setFont(gframe.config.xLabelFont);
       text.setBackground(gframe.config.bgColor);
       text.setText(gframe.config.xLabel);
       if (gframe.config.isXLabelRotate) {
          text.setRotation(90);

          text.draw(g, gframe.config.xLabelPoint.x + text.getRWidth(g),
             gframe.config.xLabelPoint.y+(int)(text.getRHeight(g)/2),
             TextLine.LEFT);

       } else {
          text.draw(g, gframe.config.xLabelPoint.x, 
                       gframe.config.xLabelPoint.y+text.getHeight(g),
                       TextLine.CENTER);
       }
       text.setRotation(0);

       text.setFont(gframe.config.yLabelFont);
       text.setText(gframe.config.yLabel);
       if (gframe.config.isYLabelRotate) {
          text.setRotation(90);

          text.draw(g, gframe.config.yLabelPoint.x + text.getRWidth(g),
             gframe.config.yLabelPoint.y+(int)(text.getRHeight(g)/2),
             TextLine.LEFT);

       } else {
          text.draw(g, gframe.config.yLabelPoint.x, 
                       gframe.config.yLabelPoint.y+text.getHeight(g),
                       TextLine.CENTER);
       }
       text.setRotation(0);

       if (isLabelMove > 0) {
         g.setColor(Color.lightGray);
         g.fillRect(gframe.config.xLabelPoint.x, 
                            gframe.config.xLabelPoint.y, 20, 20);
         g.fillRect(gframe.config.yLabelPoint.x, 
                            gframe.config.yLabelPoint.y, 20, 20);
       }

       g.setFont(saved);
    }

    // legend drawing
    if (gframe.config.display_legend) {
        calcLegendArea();

        int x = gframe.config.legendRect.x;
        int y = gframe.config.legendRect.y;
        int w = gframe.config.legendRect.width;
        int h = gframe.config.legendRect.height;

        // draw boundary rectangle
        g.setColor(gframe.config.bgColor);
        g.fillRect(x, y, w, h);
        g.setColor(Color.black);
        g.drawRect(x, y, w, h);

        int f = (int)(resizeFactor * 2.0f);
        if (f < 1)
            f = 1;
        if (f > 4)
            f = 4;
        int f2 = 2 * f;

        int gap = h / gframe.dataSetVector.size();
        int mpos = x + 20;                            // marker position
        FontMetrics fm = g.getFontMetrics();

        for (int i=0; i<gframe.dataSetVector.size(); i++) {
            DataSet1D data = (DataSet1D)gframe.dataSetVector.elementAt(i);

            int pos = y + i*gap + gap/2;                // line position
            int lblp = y + i*gap + fm.getMaxAscent();   // label position

            g.setColor(data.getLineColor());
            g.drawLine(x+5, pos, x+35, pos);

            switch (data.getMarkerType()) {
              case MarkerChooser.CIRCLEMARKER:
                g.drawOval(mpos-f, pos-f, f2, f2);
                break;

              case MarkerChooser.SQUAREMARKER:
                g.drawRect(mpos-f, pos-f, f2, f2);
                break;

              case MarkerChooser.DIAMONDMARKER:
                g.drawLine(mpos, pos-f, mpos-f, pos);
                g.drawLine(mpos-f, pos, mpos, pos+f);
                g.drawLine(mpos, pos+f, mpos+f, pos);
                g.drawLine(mpos+f, pos, mpos, pos-f);
                break;

              case MarkerChooser.CIRCLEFILLMARKER:
                g.fillOval(mpos-f, pos-f, f2, f2);
                break;

              case MarkerChooser.SQUAREFILLMARKER:
                g.fillRect(mpos-f, pos-f, f2, f2);
                break;

              case MarkerChooser.DIAMONDFILLMARKER:
                int[] xpts = {mpos,  mpos-f, mpos, mpos+f };
                int[] ypts = {pos-f, pos, pos+f, pos };

                g.fillPolygon(xpts, ypts, 4);
                break;

              case MarkerChooser.CROSSMARKER:
                g.drawLine(mpos, pos-f, mpos, pos+f);
                g.drawLine(mpos-f, pos, mpos+f, pos);
                break;

              case MarkerChooser.XMARKER:
                g.drawLine(mpos-f, pos-f, mpos+f, pos+f);
                g.drawLine(mpos+f, pos-f, mpos-f, pos+f);
                break;
            }
            g.setColor(Color.black);
            g.drawString(data.getTitle(), x + 40, lblp);
        }

        if (isLegendMove) {
            g.setColor(Color.lightGray);
            g.fillRect(legendCenter.x, legendCenter.y, 
                              legendCenter.width, legendCenter.height);
        }
    }

    // draw Axis
    axis.setAxisBounds(pda);
    axis.setAxisMinMax(minmaxArr);
    axis.paint(g);


    // display time calue
    if (gframe.config.display_time) {
       FontMetrics fm = g.getFontMetrics(g.getFont());
       g.setColor(Color.black);
       g.drawString("Time: " + 
          SVUIC.formatFloat(gframe.timeArr[gframe.currentTime], 3, 2, 
                              SVUIC.ALGEBRAIC_FORMAT),
                              5, fm.getMaxAscent()+10);
    }

    if (isPlotAreaResize) {
       for (int i=0; i<2; i++) {
          g.setColor(Color.lightGray);
          g.fillRect(pdaCorners[i].x, pdaCorners[i].y, 
                         pdaCorners[i].width, pdaCorners[i].height);
        }
    }
  }

  ///////////////////////////////////////////////////////////
  // draw point data  
  ///////////////////////////////////////////////////////////
  private void drawPData(Graphics g) {
    Dimension d;
    int totPoints = 0;
    boolean errorCodes = false;

    int f = (int)(resizeFactor * 2.0f);
    if (f < 1)
        f = 1;
    if (f > 4)
        f = 4;
    int f2 = 2 * f;
    int linewid = Math.round(resizeFactor * 1.2f);

    if (printing)
        d = new Dimension(printwidth, printheight);
    else
        d = getSize();

    g.translate(pda.x, pda.y);

    float xrange = minmaxArr[1] - minmaxArr[0];
    float yrange = minmaxArr[3] - minmaxArr[2];
    if (xrange == 0.0f) xrange = 1.0f;
    if (yrange == 0.0f) yrange = 1.0f;

    float xscale = (float)pda.width / xrange;
    float yscale = (float)pda.height / yrange;

    int xpadout = (int)(minmaxArr[0] * xscale);
    int ypadout = (int)(minmaxArr[2] * yscale);

    for (int i=0; i<gframe.dataSetVector.size(); i++) {
        DataSet1D data = (DataSet1D)gframe.dataSetVector.elementAt(i);
        for (int j=0; j<data.getTimeArr().length; j++) {
            if (data.getTimeValue(j) == gframe.timeArr[gframe.currentTime]) {
                if (((TimeData1D)(data.timeData[j])).getErrorArr() != null) 
                    errorCodes = true;

                int points = data.timeData[j].getnoOfPoints();
                if (totPoints < points)
                    totPoints = points;

                float xArr[] = data.timeData[j].getxArr();
                float yArr[] = data.timeData[j].getyArr();
                int xx[] = new int[xArr.length];
                int yy[] = new int[yArr.length];
                float eArr[] = null;
                int err[] = null;

                if (errorCodes) {
                   eArr = ((TimeData1D)(data.timeData[j])).getErrorArr();
                   err = new int[eArr.length];
                }

                // calculate position of the data point to be displayed
                for (int k=0; k<xArr.length; k++) {
                    xx[k] = (int)(xArr[k] * xscale) - xpadout;
                    yy[k] = pda.height - (int)(yArr[k] * yscale) + ypadout;
                }

                if (errorCodes) {
                   for (int k=0; k<xArr.length; k++) 
                      err[k] = (int)(eArr[k] * yscale);
                }

                for (int k=0; k<points; k++) {
                    if (gframe.config.amr || gframe.config.color_by_info) {
                        if (((TimeData1D)(data.timeData[j])).colorArr != null)
                            g.setColor(((TimeData1D)(data.timeData[j])).colorArr[k]);
                        else {
			  // one more opportunity
                            if (gframe.config.amr) {
                               ((TimeData1D)(data.timeData[j])).calcAMRColor(gframe.colorModel, gframe.rcolors, gframe.gcolors, gframe.bcolors, gframe.hcolors, gframe.scolors, gframe.lcolors);
                            } else {
                              ((TimeData1D)(data.timeData[j])).calcColorByInfo(gframe.colorModel, gframe.rcolors, gframe.gcolors, gframe.bcolors, gframe.hcolors, gframe.scolors, gframe.lcolors);
                            }
                            if (((TimeData1D)(data.timeData[j])).colorArr != null)
                               g.setColor(((TimeData1D)(data.timeData[j])).colorArr[k]);
                            else
                               g.setColor(data.getLineColor());

                        }
                    } else
                        g.setColor(data.getLineColor());

                    if (errorCodes) {
                       SVUIC.drawLine(g, xx[k], yy[k]-err[k], 
                            xx[k], yy[k]+err[k], linewid, g.getColor());
                       SVUIC.drawLine(g, xx[k]-1, yy[k]-err[k], 
                            xx[k]+1, yy[k]-err[k], linewid, g.getColor());
                       SVUIC.drawLine(g, xx[k]-1, yy[k]+err[k], 
                            xx[k]+1, yy[k]+err[k], linewid, g.getColor());
                    }

                    if (gframe.config.connect_point && k != (points-1)) 
                        SVUIC.drawLine(g, xx[k], yy[k], 
                            xx[k+1], yy[k+1], linewid, g.getColor());

                    switch (data.getMarkerType()) {
                    case MarkerChooser.CIRCLEMARKER:
                        g.drawOval(xx[k]-f, yy[k]-f, f2, f2);
                        break;

                    case MarkerChooser.SQUAREMARKER:
                        g.drawRect(xx[k]-f, yy[k]-f, f2, f2);
                        break;

                    case MarkerChooser.DIAMONDMARKER:
                        g.drawLine(xx[k], yy[k]-f, xx[k]-f, yy[k]);
                        g.drawLine(xx[k]-f, yy[k], xx[k], yy[k]+f);
                        g.drawLine(xx[k], yy[k]+f, xx[k]+f, yy[k]);
                        g.drawLine(xx[k]+f, yy[k], xx[k], yy[k]-f);
                        break;

                    case MarkerChooser.CIRCLEFILLMARKER:
                        g.fillOval(xx[k]-f, yy[k]-f, f2, f2);
                        break;

                    case MarkerChooser.SQUAREFILLMARKER:
                        g.fillRect(xx[k]-f, yy[k]-f, f2, f2);
                        break;

                    case MarkerChooser.DIAMONDFILLMARKER:
		        int[] xpts = { xx[k], xx[k]-f, xx[k], xx[k]+f };
                        int[] ypts = { yy[k]-f, yy[k], yy[k]+f, yy[k] };
                        g.fillPolygon(xpts, ypts, 4);
                        break;

                    case MarkerChooser.CROSSMARKER:
                        g.drawLine(xx[k], yy[k]-f, xx[k], yy[k]+f);
                        g.drawLine(xx[k]-f, yy[k], xx[k]+f, yy[k]);
                        break;

                    case MarkerChooser.XMARKER:
                        g.drawLine(xx[k]-f, yy[k]-f, xx[k]+f, yy[k]+f);
                        g.drawLine(xx[k]+f, yy[k]-f, xx[k]-f, yy[k]+f);
                        break;
                    }

                    if (drawClosest && xArr[k] == closest[0] 
                                              && yArr[k] == closest[1]) {
                        Color color = g.getColor();
                        g.setColor(Color.black);
                        g.fillOval(xx[k]-f, yy[k]-f, f2, f2);

                        displayStr = displayStr + "), "+(k+1) + "th: (" + 
        SVUIC.formatFloat(closest[0], xsig, xprec, SVUIC.ALGEBRAIC_FORMAT) 
        + ", " + 
        SVUIC.formatFloat(closest[1], xsig, xprec, SVUIC.ALGEBRAIC_FORMAT) 
        + ")";
                        gframe.monitor.display(displayStr, false);
                        g.setColor(color);
                    }
                }
            }
        }
    }

    g.translate(-pda.x, -pda.y);

    g.setColor(gframe.config.bgColor);
    g.fillRect(0, 0, d.width, pda.y-5);
    g.fillRect(0, 0, pda.x-5, d.height);
    g.fillRect(0, pda.y+pda.height+10, d.width, pda.y+pda.height+10);
    g.fillRect(pda.x+pda.width+10, 0, pda.x+pda.width+10, d.height);




    if (drawClosest)
        drawClosest = false;
    else
        gframe.monitor.display((gframe.currentTime+1) + " / " + 
              gframe.noOfTime + "   ( " + totPoints + " )", false);
  }


  public void keyTyped(KeyEvent e) {
    // NOTHING TO DO
  }

  public void keyReleased(KeyEvent e) {
    // NOTHING TO DO
  }


  // key pressed from a-65  to z-90
  public void keyPressed(KeyEvent e) {
    //System.out.println(e.getKeyCode());

    if (e.isControlDown()) {   

        switch(e.getKeyCode()) {
           case 79:                       // ctrl-o zoom out
	      svServer.fmanager.action1d(gframe, CSPlane.ZOOMOUT, 
                                                              null, true);
              break;
           case 65:                       // ctrl-a animation
              svServer.fmanager.action1d(gframe, CSPlane.ANIMATE, 
                                                              null, true);
              break;

           case 83:                       // ctrl-s stop
              String[] arg = new String[1];
              arg[0] = String.valueOf(gframe.currentTime);
              svServer.fmanager.action1d(gframe, CSPlane.STOP, 
                                                              arg, true);
              break;

           case 78:                       // ctrl-n next
              svServer.fmanager.action1d(gframe, CSPlane.NEXT, 
                                                              null, true);
              break;

           case 80:                      // ctrl-p previous
              svServer.fmanager.action1d(gframe, CSPlane.PREVIOUS, 
                                                              null, true);
              break;

           case 81:                      // ctrl-q kill window
              svServer.fmanager.action1d(gframe, CSPlane.QUIT, 
                                                              null, true);
              break;

           case 71:                      // ctrl-g go specific time
              gframe.doWork(CSPlane.GOTO, null);
              break;

           case 82:                      // ctrl-r setting panel

              break;
        }

      } else if (e.isAltDown()) {     // user filters short key
         Vector menuVec = svServer.menuHandler.menuVec;
         for (Enumeration e2=menuVec.elements(); e2.hasMoreElements();) {
            MenuInfo info = (MenuInfo)e2.nextElement();
            if (info.shortkey == e.getKeyCode()) {
               gframe.doFilterAction(info.name);
               break;
            }
         } 
      }
  }


  public void mouseDragged(MouseEvent e)  {

     int x = e.getX();
     int y = e.getY();

     Rectangle rect;

     if (isPlotAreaResize) {
       switch (selectedCorner) {
         case 0:
             rect = 
                new Rectangle(x, y, pda.x+pda.width-x, pda.y+pda.height- y);
             if (rect.width > 0 && rect.height > 0)
                 pda = rect;
             break;

         case 1:
             int hgt = pda.y - y;
             int wid = pda.x - x;

             if (hgt < 0)
                 hgt *= -1;
             if (wid < 0)
                 wid *= -1;

             rect = new Rectangle(pda.x, pda.y, wid, hgt);

             if (rect.width > 0 && rect.height > 0)
                 pda = rect;
             break;

         default:
             return;
       }
       change_pdaCorners();
       repaint();
       return;
     }

     if (isLegendMove) {
         Rectangle r1 = gframe.config.legendRect;
         Rectangle r2 = new Rectangle(x-r1.width/2, y-r1.height/2, 
                                                     r1.width, r1.height);
         int midx = Math.round((float)r2.x + (float)r2.width / 2.0f);
         int midy = Math.round((float)r2.y + (float)r2.height / 2.0f);

         legendCenter = new Rectangle(midx-10, midy-10, 20, 20);
         repaint();
         return;
     }

     
     if (isLabelMove == 1) {
         gframe.config.xLabelPoint.x = x;
         gframe.config.xLabelPoint.y = y;
         repaint();
         return;
     } else if (isLabelMove == 2) {
         gframe.config.yLabelPoint.x = x;
         gframe.config.yLabelPoint.y = y;
         repaint();
         return;
     }

     if (drawingmode) {
       if (svServer.shapes.getShapeType() < 0) {

          if (hotspot != null) {
            hotspot.setPoint(new Point(x, y));
            hotspot.notifyObservers();
            repaint();
          }
          return;
       }

       if (svServer.shapes.getShapeType() == ShapeProperties.MOUSE) {
          mPointVec.addElement(new Point(x, y));
          isDragged = true;
          repaint();
          return;
       }

     } else if (zoomMode == 3) return;

     isDragged = true;
     dragEndX = x;
     dragEndY = y;

     if (dragEndX > (pda.x+pda.width)) 
         dragEndX = pda.x+pda.width;
     if (dragEndY > (pda.y+pda.height)) 
         dragEndY = pda.y+pda.height;
     if (dragEndX < pda.x) dragEndX = pda.x;
     if (dragEndY < pda.y) dragEndY = pda.y;

     if (dragEndX < dragStartX) inversedX = true;
     else if (inversedX) inversedX = false;

     if (dragEndY < dragStartY) inversedY = true;
     else if (inversedY) inversedY = false;
     repaint();
  }

  public void mouseMoved(MouseEvent evt) {
     
     if (drawingmode && svServer.shapes.getShapeType() < 0) {
        int x = evt.getX();
        int y = evt.getY();

        for (int i = 0; i < dh.drawingVec.size(); i++) {
           ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(i);
           s.setFocus(false);
        }

        for (int i = dh.drawingVec.size()-1; i >= 0; i--) {
           ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(i);
             
           if (s.contains(x, y)) { 
              s.setFocus(true);
              break;
           }
	}
        repaint();
     }
  }

  // overrided function
  public void processMouseEvent(MouseEvent evt) {

    if (evt.isPopupTrigger()) { 
       if (drawingmode) {
          int x = evt.getX();
          int y = evt.getY();

          wcutmi.setEnabled(false);
          wcopymi.setEnabled(false);
          wpastemi.setEnabled(false);
          wfillmi.setEnabled(false);
          wlinecmi.setEnabled(false);
          wlinepmi.setEnabled(false);
          wfillcmi.setEnabled(false);
          wtextcmi.setEnabled(false);
          wtextfmi.setEnabled(false);

          if (cuttedObject != null)
             wpastemi.setEnabled(true);
          
          for (int i = 0; i < dh.drawingVec.size(); i++) {
             ShapeObject s = (ShapeObject)dh.drawingVec.elementAt(i);

             if (s.isFocused()) {
	        switch (s.getType()) {
                   case ShapeProperties.LINE:
                   case ShapeProperties.MULTILINE:
                   case ShapeProperties.BEZIER:
                   case ShapeProperties.MOUSE:
                      wcutmi.setEnabled(true);
                      wcopymi.setEnabled(true);
                      wlinecmi.setEnabled(true);
                      wlinepmi.setEnabled(true);
                      break;
                   case ShapeProperties.RECTANGLE:
                   case ShapeProperties.POLYGON:
                   case ShapeProperties.OVAL:
                      wcutmi.setEnabled(true);
                      wcopymi.setEnabled(true);
                      wfillmi.setEnabled(true);
                      wfillcmi.setEnabled(true);
                      wlinecmi.setEnabled(true);
                      break;

                   case ShapeProperties.TEXT:
                      wcutmi.setEnabled(true);
                      wcopymi.setEnabled(true);
                      wtextcmi.setEnabled(true);
                      wtextfmi.setEnabled(true);
                      break;
	        }
                break;
             }
          }
          wpopup.show(evt.getComponent(), evt.getX(), evt.getY());
       }
    }

     if (evt.getID() == MouseEvent.MOUSE_PRESSED) { 
        mPressed(evt);
     } else if (evt.getID() == MouseEvent.MOUSE_RELEASED) { 
        mReleased(evt);
     }

    super.processMouseEvent(evt);
  }
 
  /////////////////////////////////////////////////////////////
  // if mouse pressed
  /////////////////////////////////////////////////////////////
  private void mPressed(MouseEvent e) {
    int x = e.getX();
    int y = e.getY();

    if (isPlotAreaResize) {
        for (int i=0; i<2; i++) {
            if (pdaCorners[i].contains(x, y)) {
                selectedCorner = i;
                return;
            }
        }
        isPlotAreaResize = false;
    }

    if (isLegendMove) {
        if (legendCenter.contains(x, y))
            return;
        isLegendMove = false;
    }

    if (isLabelMove > 0) {
         Rectangle xrect = new Rectangle(gframe.config.xLabelPoint.x, 
                            gframe.config.xLabelPoint.y, 20, 20);
         Rectangle yrect = new Rectangle(gframe.config.yLabelPoint.x, 
                            gframe.config.yLabelPoint.y, 20, 20);

        if (xrect.contains(x, y)) {
           isLabelMove = 1;
           return;
        }
 
        if (yrect.contains(x, y)) {
           isLabelMove = 2;
           return;
        }

        isLabelMove = -1;
    }

    if (drawingmode) {

      if (svServer.shapes.getShapeType() < 0) {
         dh.hotspotManager.isAt(new Point(x, y));
         int pos = dh.hotspotManager.HSPosition();

         if (pos > 0) {
            String arg[] = new String[1];
            arg[0] = String.valueOf(dh.hotspotManager.HSPosition());
            svServer.fmanager.action1d(gframe, CommandSet.WHSCREATE, 
                                                          arg, true);
         }
         return;

      } else if (multiShape == null) {

         switch(svServer.shapes.getShapeType()) {

            case ShapeProperties.MULTILINE:
              multiShape = new MLineShape(ShapeProperties.MULTILINE,  
                                 x, y, 
                                 svServer.shapes.getLineColor(),
                                 svServer.shapes.getLineType(),
                                 svServer.shapes.getLineWidth(),
                                 svServer.shapes.getDashSize() );
              break;

            case ShapeProperties.BEZIER:
              multiShape = new BLineShape(ShapeProperties.BEZIER,  
                                 x, y, 
                                 svServer.shapes.getLineColor(),
                                 svServer.shapes.getLineType(),
                                 svServer.shapes.getLineWidth(),
                                 svServer.shapes.getDashSize() );
              break;

            case ShapeProperties.MOUSE:
              multiShape = new RLineShape(ShapeProperties.MOUSE,  
                                 svServer.shapes.getLineColor(),
                                 svServer.shapes.getLineType(),
                                 svServer.shapes.getLineWidth(),
                                 svServer.shapes.getDashSize() );
              mPointVec = new Vector();
              mPointVec.addElement(new Point(x, y));
              break;
            case ShapeProperties.POLYGON:
              multiShape = new PolygonShape(ShapeProperties.POLYGON,  
                                 svServer.shapes.getLineColor(),
                                 svServer.shapes.getFillColor(),
                                 "false");
              mPointVec = new Vector();
              mPointVec.addElement(new Point(x, y));
              break;
         }

         if (multiShape != null) {
            String arg[] = new String[1];
            arg[0] = multiShape.makeString();
            svServer.fmanager.action1d(gframe, CommandSet.WCREATE, 
                                                         arg, true);
         }
       }
              

    } else if (zoomMode == 3) return;

    dragStartX = e.getX();
    dragStartY = e.getY();

    if (dragStartX > (pda.x+pda.width))  dragStartX = pda.x+pda.width;
    if (dragStartY > (pda.y+pda.height)) dragStartY = pda.y+pda.height;
    if (dragStartX < pda.x) dragStartX = pda.x;
    if (dragStartY < pda.y) dragStartY = pda.y;
  }

  ///////////////////////////////////////////////////////////// 
  // if mouse released
  /////////////////////////////////////////////////////////////
  private void mReleased(MouseEvent e) {
    int x = e.getX();
    int y = e.getY();
    int sx, sy, ex, ey;

    if (isPlotAreaResize) {
       String args[] = { String.valueOf(pda.x), 
                         String.valueOf(pda.y), 
                         String.valueOf(pda.width), 
                         String.valueOf(pda.height) };
       svServer.fmanager.action1d(gframe, CSPlane.RESIZEPDA, args, true);
    }

    if (isLegendMove) {
       String args[] = { String.valueOf(gframe.config.legendRect.x), 
                         String.valueOf(gframe.config.legendRect.y), 
                         String.valueOf(gframe.config.legendRect.width), 
                         String.valueOf(gframe.config.legendRect.height) };
       svServer.fmanager.action1d(gframe, CSPlane.LEGENDMOVE, args, true);
    }

    if (isLabelMove > 0) {
       String args[] = { String.valueOf(gframe.config.xLabelPoint.x), 
                         String.valueOf(gframe.config.xLabelPoint.y), 
                         String.valueOf(gframe.config.yLabelPoint.x), 
                         String.valueOf(gframe.config.yLabelPoint.y) };
       svServer.fmanager.action1d(gframe, CSPlane.LABELMOVE, args, true);
    }

    if (drawingmode) {
      if (hotspot != null) {
         String arg[] = new String[2];
         arg[0] = String.valueOf(x);
         arg[1] = String.valueOf(y);
         svServer.fmanager.action1d(gframe, CommandSet.WHSMOVE, 
                                                          arg, true);
       }

       if (e.getClickCount() > 1) {

          if (svServer.shapes.getShapeType() == ShapeProperties.POLYGON) {
             isDragged = false;
             int sz = mPointVec.size();
             String arg[] = new String[sz*2 + 2];

             arg[0] = String.valueOf(ShapeProperties.POLYGON);
             arg[1] = String.valueOf(sz);

             int idx = 2;
             for (int i=0; i<sz; i++) {
                Point p = (Point)mPointVec.elementAt(i);
                arg[idx] = String.valueOf(p.x); idx++;
                arg[idx] = String.valueOf(p.y); idx++;
              }

              svServer.fmanager.action1d(gframe, 
                        CommandSet.WADDPOINTS, arg, true);

              mPointVec = null;
              multiShape = null;
              svServer.dtool.clearAllToggleButtons();            
              return;
	  }

          multiShape = null;
          svServer.dtool.clearAllToggleButtons();  
       }

       if (svServer.shapes.getShapeType() >= 0 && isDragged) {
          if (x > dragStartX) {  
             sx = dragStartX;
             ex = x;   
          } else {
             ex = dragStartX;
             sx = x;  
          }
          if (y > dragStartY) {  
             sy = dragStartY;
             ey = y;   
          } else {
             ey = dragStartY;
             sy = y;  
          }
         
          isDragged = false;

          ShapeObject created = null;
          String arg[];

          switch(svServer.shapes.getShapeType()) {
            case ShapeProperties.LINE:
              created = new LineShape(ShapeProperties.LINE, 
                                 dragStartX, dragStartY, x, y, 
                                 svServer.shapes.getLineColor(),
                                 svServer.shapes.getLineType(),
                                 svServer.shapes.getLineWidth(),
                                 svServer.shapes.getDashSize() );

              break;
            case ShapeProperties.MULTILINE:
              arg = new String[3];
              arg[0] = String.valueOf(ShapeProperties.MULTILINE);
              arg[1] = String.valueOf(x);
              arg[2] = String.valueOf(y);
              svServer.fmanager.action1d(gframe, 
                       CommandSet.WADDPOINT, arg, true);
              return;

            case ShapeProperties.BEZIER:
              arg = new String[3];
              arg[0] = String.valueOf(ShapeProperties.BEZIER);
              arg[1] = String.valueOf(x);
              arg[2] = String.valueOf(y);
              svServer.fmanager.action1d(gframe, 
                       CommandSet.WADDPOINT, arg, true);
              return;

            case ShapeProperties.MOUSE:
              isDragged = false;
              int sz = mPointVec.size();
              arg = new String[sz*2 + 2];

              arg[0] = String.valueOf(ShapeProperties.MOUSE);
              arg[1] = String.valueOf(sz);

              int idx = 2;
              for (int i=0; i<sz; i++) {
                Point p = (Point)mPointVec.elementAt(i);
                arg[idx] = String.valueOf(p.x); idx++;
                arg[idx] = String.valueOf(p.y); idx++;
              }

              svServer.fmanager.action1d(gframe, 
                        CommandSet.WADDPOINTS, arg, true);

              multiShape = null;
              svServer.dtool.clearAllToggleButtons();
              mPointVec = null;
              return;

            case ShapeProperties.RECTANGLE:
              created = new RectangleShape(ShapeProperties.RECTANGLE, 
                                 sx, sy, ex-sx, ey-sy);
              ((RectangleShape)created).setLineColor(svServer.shapes.getLineColor());
              ((RectangleShape)created).setFillColor(svServer.shapes.getFillColor());
              break;
            case ShapeProperties.POLYGON:
              mPointVec.addElement(new Point(x, y));
              isDragged = true;
              repaint();
              return;

            case ShapeProperties.OVAL:
              created = new OvalShape(ShapeProperties.OVAL, 
                                 sx, sy, ex-sx, ey-sy);
              ((OvalShape)created).setLineColor(svServer.shapes.getLineColor());
              ((OvalShape)created).setFillColor(svServer.shapes.getFillColor());
              break;
            case ShapeProperties.TEXT:
              String result = 
                    JOptionPane.showInputDialog(gframe, "Enter text:");

              if (result != null) {
                 created = new TextShape(ShapeProperties.TEXT, 
                                                 sx, sy, ex-sx, ey-sy);

                 ((TextShape)created).setFont(gframe.svServer.shapes.getTextFont());
                 ((TextShape)created).setTextColor(gframe.svServer.shapes.getTextColor());
                 ((TextShape)created).setLabel(result);
	      }
              break;

            case ShapeProperties.IMAGE:
              FileDialog chooser = new FileDialog(gframe,
                                       "Select Image", FileDialog.LOAD);
              chooser.show();
              svServer.dtool.clearAllToggleButtons();
              String input = chooser.getDirectory() + chooser.getFile();
              if (input == null) return;

              Image img;
              try {
                img = getToolkit().getImage(input);
              } catch (Exception E) { 
                System.out.println("Can't load the image ...");
                return;
              }

              if (img != null) {
                created = new ImageShape(ShapeProperties.IMAGE, 
                                       sx, sy, ex-sx, ey-sy);
                dh.createDrawingObject(created.makeString(), img);
                repaint();
              } else { 
                System.out.println("Can't load the image ...");
                return;
              }

              return;
          }

          if (created != null) {
             arg = new String[1];
             arg[0] = created.makeString();
             svServer.fmanager.action1d(gframe, CommandSet.WCREATE, 
                                                          arg, true);
          }
          svServer.dtool.clearAllToggleButtons();
       }
              
       return;
    }

    float xmin = minmaxArr[0];
    float xmax = minmaxArr[1];
    float ymin = minmaxArr[2];
    float ymax = minmaxArr[3];

    if (isDragged && zoomMode != 3) {

        isDragged = false;
        gframe.zoomoutMI.setEnabled(true);

        if (zoomMode == 0 || zoomMode == 1) {
            float zoomScaleX = (float)pda.width / (xmax - xmin);

            if (inversedX)
                x = dragEndX - pda.x;
            else
                x = dragStartX - pda.x;

            float zoomStartX = (float)x / zoomScaleX + xmin;

            if (inversedX)
                x = dragStartX - pda.x;
            else
                x = dragEndX - pda.x;

            float zoomEndX = (float)x / zoomScaleX + xmin;

            zoomFactorX = zoomFactorX * (xmax - xmin) / (zoomEndX - zoomStartX);

            minmaxArr[0] = zoomStartX;
            minmaxArr[1] = zoomEndX;
        }

        if (zoomMode == 0 || zoomMode == 2) {
            float zoomScaleY = (float)pda.height / (ymax - ymin);

            if (inversedY)
                y = dragEndY - pda.y;
            else
                y = dragStartY - pda.y;

            float zoomEndY = (float)(pda.height - y) / zoomScaleY + ymin;

            if (inversedY)
                y = dragStartY - pda.y;
            else
                y = dragEndY - pda.y;

            float zommStartY = (float)(pda.height - y) / zoomScaleY + ymin;

            zoomFactorY = zoomFactorY * (ymax - ymin) / (zoomEndY - zommStartY);
            minmaxArr[2] = zommStartY;
            minmaxArr[3] = zoomEndY;
        }

        inversedX = false;
        inversedY = false;

        String args[] = { String.valueOf(minmaxArr[0]), 
                          String.valueOf(minmaxArr[1]), 
                          String.valueOf(minmaxArr[2]), 
                          String.valueOf(minmaxArr[3]), 
                          String.valueOf(zoomFactorX), 
                          String.valueOf(zoomFactorY), 
                          String.valueOf(zoomMode) };
        svserver.fmanager.action1d(gframe, CSPlane.ZOOMING, args, true);
        return;
    }

    if (svserver.mergeHandler.mergeMode) {
        gframe.mergeAction(e.isControlDown());
        return;
    }

    x = e.getX() - pda.x + 5;
    y = e.getY() - pda.y + 5;

    Rectangle rect = new Rectangle(pda.width+5, pda.height+5);

    if (rect.contains(x, y)) {
        float scalex = (float)pda.width / (xmax-xmin);
        float scaley = (float)pda.height / (ymax-ymin);

        float clickedx = (float)x/scalex + xmin;
        float clickedy = (float)(pda.height-y)/scaley + ymin;

        closest = getClosestPoint(clickedx, clickedy);

        drawClosest = true;

        displayStr =  "(" + 
   SVUIC.formatFloat(clickedx, xsig, xprec, SVUIC.ALGEBRAIC_FORMAT) + ", " + 
   SVUIC.formatFloat(clickedy, ysig, yprec, SVUIC.ALGEBRAIC_FORMAT);
        repaint();
    }
  }

  private void calcLegendArea() {
    int size = gframe.dataSetVector.size();

    FontMetrics fm = offScrGc.getFontMetrics();
    Dimension d = getSize();

    int hgt = (fm.getHeight()+10) * size;
    int wid = 0;

    Enumeration e;
    for (e = gframe.dataSetVector.elements(); e.hasMoreElements(); ) {
        DataSet1D data = (DataSet1D)e.nextElement();

        int w = fm.stringWidth(data.getTitle());
        if (wid < w)
            wid = w;
    }

    wid += 50;

    if (legendCenter == null) {
        gframe.config.legendRect = 
             new Rectangle(d.width-wid-10, 10, wid, hgt);

        int midx = Math.round((float)gframe.config.legendRect.x + 
                           (float)gframe.config.legendRect.width / 2.0f);
        int midy = Math.round((float)gframe.config.legendRect.y + 
                           (float)gframe.config.legendRect.height / 2.0f);
        legendCenter = new Rectangle(midx-10, midy-10, 20, 20);
        return;
    }

    gframe.config.legendRect = 
      new Rectangle(legendCenter.x-wid/2, legendCenter.y-hgt/2, wid, hgt);
  }



} // end of class











